from enum import Enum


class LogFormat(str, Enum):
    """Presets for logging formats that can be used in :func:`set_log`.

    ``{color_start}`` and ``{color_end}`` are used to add colors to parts of the log message.
    If they are not used, the whole log message will be colored.
    """

    full_color = "[%(levelname)s] %(message)s"
    full_color_time = "[%(asctime)s] %(levelname)s: %(message)s"
    color_level = "[{color_start}%(levelname)s{color_end}] %(message)s"
    default = color_level

    def __str__(self):
        return self.value



"""Embed templates that can be used to send messages.
These functions will generate embeds and send them to the desired target.

Example
-------
.. code-block:: python

    from ezcord import Bot, emb

    bot = Bot()

    @bot.slash_command()
    async def hey(ctx):
        await emb.success(ctx, "Success!")
"""

from __future__ import annotations

import discord

from .internal.embed_templates import load_embed, save_embeds


def set_embed_templates(
    *,
    error_embed: discord.Embed | None = None,
    success_embed: discord.Embed | None = None,
    warn_embed: discord.Embed | None = None,
    info_embed: discord.Embed | None = None,
    **kwargs: discord.Embed,
):
    """Override the default embeds with custom ones.

    This must be called before the first embed template is used. The description
    of the embeds will be replaced with the given text.

    Parameters
    ----------
    error_embed:
        The embed to use for error messages.
    success_embed:
        The embed to use for success messages.
    warn_embed:
        The embed to use for warning messages.
    info_embed:
        The embed to use for info messages.
    **kwargs:
        Additional embed templates. Can be used with :func:`send`.

    Example
    -------
    .. code-block:: python

        from ezcord import emb

        embed = discord.Embed(
            title="Error",
            color=discord.Color.orange()
        )
        emb.set_embed_templates(error_embed=embed)
    """
    save_embeds(
        error_embed=error_embed,
        success_embed=success_embed,
        warn_embed=warn_embed,
        info_embed=info_embed,
        **kwargs,
    )


async def _send_embed(
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    embed: discord.Embed,
    ephemeral: bool = True,
    **kwargs,
):
    """Respond to an interaction or send an embed to a target.

    If the interaction has already been responded to,
    the message will be sent as a followup.

    Parameters
    ----------
    target:
        The application context or the interaction to send the message to.
    embed:
        The embed to send.
    ephemeral:
        Whether the message should be ephemeral. Defaults to ``True``.
    """
    if isinstance(target, discord.ApplicationContext) or isinstance(target, discord.Interaction):
        if not target.response.is_done():
            await target.response.send_message(embed=embed, ephemeral=ephemeral, **kwargs)
        else:
            await target.followup.send(embed=embed, ephemeral=ephemeral, **kwargs)
    else:
        await target.send(embed=embed, **kwargs)


async def _process_message(
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    embed: discord.Embed,
    txt: str,
    ephemeral: bool,
    *,
    title: str | None = None,
    **kwargs,
):
    """Adds embed attributes to the embed before sending it.

    Parameters
    ----------
    target:
        The target to send the message to.
    txt:
        The text to send.
    ephemeral:
        Whether the message should be ephemeral.
    """
    embed.description = txt
    if title is not None:
        embed.title = title

    await _send_embed(target, embed, ephemeral, **kwargs)


async def error(
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    txt: str,
    ephemeral: bool = True,
    **kwargs,
):
    """Send an error message. By default, this is a red embed.

    Parameters
    ----------
    target:
        The target to send the message to.
    txt:
        The text to send.
    ephemeral:
        Whether the message should be ephemeral. Defaults to ``True``.
    """
    embed = load_embed("error_embed")
    await _process_message(target, embed, txt, ephemeral, **kwargs)


async def success(
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    txt: str,
    ephemeral: bool = True,
    **kwargs,
):
    """Send a success message. By default, this is a green embed.

    Parameters
    ----------
    target:
        The target to send the message to.
    txt:
        The text to send.
    ephemeral:
        Whether the message should be ephemeral. Defaults to ``True``.
    """
    embed = load_embed("success_embed")
    await _process_message(target, embed, txt, ephemeral, **kwargs)


async def warn(
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    txt: str,
    ephemeral: bool = True,
    **kwargs,
):
    """Send a warning message. By default, this is an orange embed.

    Parameters
    ----------
    target:
        The target to send the message to.
    txt:
        The text to send.
    ephemeral:
        Whether the message should be ephemeral. Defaults to ``True``.
    """
    embed = load_embed("warn_embed")
    await _process_message(target, embed, txt, ephemeral, **kwargs)


async def info(
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    txt: str,
    ephemeral: bool = True,
    **kwargs,
):
    """Send an info message. By default, this is a blue embed.

    Parameters
    ----------
    target:
        The target to send the message to.
    txt:
        The text to send.
    ephemeral:
        Whether the message should be ephemeral. Defaults to ``True``.
    """
    embed = load_embed("info_embed")
    await _process_message(target, embed, txt, ephemeral, **kwargs)


async def send(
    template: str,
    target: discord.ApplicationContext | discord.Interaction | discord.abc.Messageable,
    txt: str,
    ephemeral: bool = True,
    **kwargs,
):
    """Send a custom embed template. This needs to be set up with :func:`set_embed_templates`.

    Parameters
    ----------
    template:
        The name of the template that was used in :func:`set_embed_templates`.
    target:
        The target to send the message to.
    txt:
        The text to send.
    ephemeral:
        Whether the message should be ephemeral. Defaults to ``True``.
    """
    embed = load_embed(template)
    await _process_message(target, embed, txt, ephemeral, **kwargs)



from __future__ import annotations

import logging
import re

from colorama import Fore

DEFAULT_COLOR = Fore.MAGENTA

DEFAULT_LOG_COLORS: dict[int, str] = {
    logging.DEBUG: Fore.GREEN,
    logging.INFO: Fore.CYAN,
    logging.WARNING: Fore.YELLOW,
    logging.ERROR: Fore.RED,
    logging.CRITICAL: Fore.RED,
}


def get_escape_code(color_string: str | bool) -> str:
    """Converts a color string to an ansi escape code using colorama.

    If the string already is an escape code, it will be returned.
    """
    if not isinstance(color_string, str):
        return DEFAULT_COLOR

    if color_string.startswith("\x1b["):
        return color_string

    try:
        return getattr(Fore, color_string.upper())
    except AttributeError:
        raise ValueError(
            f"{color_string} is not a valid color string. Use either the colorama library or a string like 'red'."
        )


def replace_second(string: str, substring: str, color: str) -> str:
    """Replaces custom formatters with ansi colors.

    Example
    -------
    >>> replace_second("Hello **world**!", "**", Fore.RED)
    """
    while substring in string:
        string = string.replace(substring, color, 1)
        string = string.replace(substring, Fore.RESET, 1)

    return string


def replace_dc_format(string: str, color: str | None = None) -> str:
    """Replaces Discord markdown with ansi colors."""
    if color is None:
        color = DEFAULT_COLOR
    color = get_escape_code(color)

    string = replace_second(string, "**", color)
    string = re.sub("```.*?```", "", string)  # remove codeblocks
    return string



__title__ = "ezcord"
__author__ = "tibue99"
__license__ = "MIT"
__version__ = "0.1.4"

from . import emb, logs, times, utils
from .bot import Bot
from .emb import *
from .enums import *
from .logs import *
from .times import *
from .utils import *



"""Some utility functions for Pycord and Python."""

import io
import json
from typing import Dict

import discord


def create_json_file(dictionary: Dict, filename: str = "data.json", **kwargs):
    """Create a :class:`discord.File` object from a dictionary.

    Parameters
    ----------
    dictionary:
        The dictionary to convert to a JSON file.
    filename:
        The filename to use for the JSON file.
    **kwargs:
        Keyword arguments for :class:`discord.File`.

    Returns
    -------
    :class:`discord.File`
    """
    content = json.dumps(dictionary, indent=2).encode()
    return discord.File(io.BytesIO(content), filename=filename, **kwargs)


def create_text_file(text: str, filename: str = "data.txt", **kwargs):
    """Create a :class:`discord.File` object from a string.

    Parameters
    ----------
    text:
        The string to convert to a text file.
    filename:
        The filename to use for the text file.
    **kwargs:
        Keyword arguments for :class:`discord.File`.

    Returns
    -------
    :class:`discord.File`
    """
    return discord.File(io.BytesIO(text.encode()), filename=filename, **kwargs)



"""Some logging utilities that are used for bot logs."""

from __future__ import annotations

import asyncio
import logging
import os
import sys

import aiohttp
import discord
from colorama import Fore

from .enums import LogFormat

from .internal.colors import (  # isort: skip
    DEFAULT_COLOR,
    DEFAULT_LOG_COLORS,
    get_escape_code,
    replace_dc_format,
)

DEFAULT_LOG = "ezcord"
log = logging.getLogger(DEFAULT_LOG)


def custom_log(
    key: str, message: str, *, color: str | bool = DEFAULT_COLOR, level: int = logging.INFO
):
    """Log a message with a custom log level.

    Parameters
    ----------
    key:
        The name of the custom log level.
    message:
        The message to log.
    color:
        The color to use for the log level. Defaults to ``Fore.MAGENTA``.
    level:
        The log level. Defaults to ``logging.INFO``.
    """
    color = get_escape_code(color)
    logging.getLogger(DEFAULT_LOG).log(level, message, extra={"key": key, "color": color})


def _format_log_colors(log_format: str, file: bool, final_colors: dict[int, str]) -> dict[int, str]:
    """Checks if the is sent to a file and formats the colors accordingly."""
    color_formats = {}
    if "{color_start}" in log_format and "{color_end}" in log_format:
        for level in final_colors:
            if file:
                color_formats[level] = log_format.format(color_start="", color_end="")
            else:
                color_formats[level] = log_format.format(
                    color_start=final_colors[level], color_end=Fore.RESET
                )
    else:
        for level in final_colors:
            color_formats[level] = final_colors[level] + log_format + Fore.RESET

    return color_formats


def _format_colors(colors: dict[int, str] | str | None = None) -> dict[int, str]:
    """Overwrite the default colors for the given log levels in the given format."""

    final_colors = DEFAULT_LOG_COLORS.copy()
    if colors is None:
        colors = final_colors

    if isinstance(colors, str):
        colors = get_escape_code(colors)
        for level in final_colors:
            final_colors[level] = colors
    else:
        for level in colors:
            final_colors[level] = get_escape_code(colors[level])

    return final_colors


class _ColorFormatter(logging.Formatter):
    """A logging formatter that adds colors to the output. This is used by :func:`set_log`.

    Parameters
    ----------
    file:
        Whether to log to a file.
    log_format:
        The log format.
    time_format:
        The time format.
    colors:
        Colors for the log levels.
    """

    def __init__(
        self,
        file: bool,
        log_format: str,
        time_format: str,
        dc_codeblocks: bool,
        colors: dict[int, str] | str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.file = file
        self.colors = colors
        self.dc_codeblocks = dc_codeblocks
        self.LOG_FORMAT = log_format
        self.TIME_FORMAT = time_format

    def format(self, record: logging.LogRecord):
        """Adds colors to log messages and formats them accordingly.
        Can be used with :func:`set_log`.

        Parameters
        ----------
        record:
            The log record to format.
        """
        if "color" in record.__dict__:
            colors = record.__dict__["color"]
        else:
            colors = self.colors

        color_formats = _format_colors(colors)

        if record.levelno not in color_formats:
            if "color" in record.__dict__:
                # if a custom log level is used by .custom_log()
                color_formats[record.levelno] = colors
            else:
                # if no color is set for a custom log level used by .log()
                color_formats[record.levelno] = DEFAULT_COLOR

        color_log_formats = _format_log_colors(self.LOG_FORMAT, self.file, color_formats)

        log_format = color_log_formats.get(record.levelno)
        if "key" in record.__dict__ and log_format:
            log_format = log_format.replace("%(levelname)s", record.__dict__["key"])

        current_level_color = color_formats.get(record.levelno)
        new_record = logging.makeLogRecord(record.__dict__)
        new_record.msg = replace_dc_format(new_record.msg, current_level_color)

        formatter = logging.Formatter(log_format, self.TIME_FORMAT)
        return formatter.format(new_record)


class _DiscordHandler(logging.Handler):
    """A logging handler that sends logs to a Discord webhook."""

    def __init__(
        self, webhook_url: str | None, dc_codeblocks: bool, dc_format: str, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.webhook_url = webhook_url
        self.dc_codeblocks = dc_codeblocks
        self.dc_format = dc_format

    def emit(self, record: logging.LogRecord):
        """Removes exception information and sets the log format."""
        record._exc_info_hidden, record.exc_info = record.exc_info, None
        record.exc_text = None

        if self.dc_codeblocks:
            msg = f"```ansi\n{self.format(record)}```"
        else:
            if "key" in record.__dict__:
                log_format = self.dc_format.replace("%(levelname)s", record.__dict__["key"])

                formatter = logging.Formatter(log_format)
                msg = formatter.format(record)
            else:
                msg = self.format(record)

        if self.webhook_url:
            loop = asyncio.get_event_loop()
            if loop.is_running():
                loop.create_task(_send_discord_log(self.webhook_url, record, msg))
            else:
                loop.run_until_complete(_send_discord_log(self.webhook_url, record, msg))


async def _send_discord_log(webhook_url: str, record: logging.LogRecord, msg):
    async with aiohttp.ClientSession() as session:
        webhook = discord.Webhook.from_url(webhook_url, session=session)
        name = "EzCord" if record.name == DEFAULT_LOG else record.name
        try:
            await webhook.send(
                content=msg,
                username=f"{name} Log",
            )
        except discord.HTTPException:
            log.error(
                "Error while sending log message to webhook. Please check if the URL is correct."
            )


def _discord_filter(record):
    if "webhook_sent" in record.__dict__:
        return not record.__dict__["webhook_sent"]
    return True


def set_log(
    name: str = DEFAULT_LOG,
    log_level: int = logging.DEBUG,
    *,
    file: bool = False,
    log_format: str | LogFormat = LogFormat.default,
    time_format: str = "%Y-%m-%d %H:%M:%S",
    discord_log_level: int = logging.WARNING,
    webhook_url: str | None = None,
    dc_codeblocks: bool = True,
    colors: dict[int, str] | str | None = None,
):
    """Creates a logger. If this logger already exists, it will return the existing logger.

    Parameters
    ----------
    name:
        The name of the logger.
    log_level:
        Whether to enable debug logs. Defaults to ``logging.INFO``.
    file:
        Whether to log to a file. Defaults to ``False``.
    log_format:
        The log format. Defaults to :attr:`LogFormat.default`.
    time_format:
        The time format. Defaults to ``%Y-%m-%d %H:%M:%S``.
    discord_log_level:
        The log level for discord log messages. Defaults to ``logging.WARNING``.

        .. note::
            For discord log messages, ``webhook_url`` is required.
    webhook_url:
        The discord webhook URL to send logs to. Defaults to ``None``.
    dc_codeblocks:
        Whether to use codeblocks for all Discord log messages. Defaults to ``True``.
    colors:
        A dictionary of log levels and their corresponding colors. If only one color is given,
        all log levels will be colored with that color.

        Example
        -------
        .. code-block:: python

            import logging
            from colorama import Fore
            import ezcord

            colors = {
                logging.DEBUG: "blue",
                logging.INFO: Fore.MAGENTA,
            }

            ezcord.set_log(colors=colors)

    Returns
    -------
    :class:`logging.Logger`
    """
    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    if logger.handlers:
        return logger

    handler: logging.FileHandler | logging.StreamHandler
    if file:
        if not os.path.exists("logs"):
            os.mkdir("logs")
        filename = name.split(".")[-1]
        handler = logging.FileHandler(f"logs/{filename}.log", mode="w", encoding="utf-8")
    else:
        handler = logging.StreamHandler(sys.stdout)

    color_formatter = _ColorFormatter(file, log_format, time_format, dc_codeblocks, colors)

    handler.setFormatter(color_formatter)
    handler.setLevel(log_level)
    logger.addHandler(handler)

    dc_format = "**%(levelname)s:** %(message)s"
    discord_handler = _DiscordHandler(webhook_url, dc_codeblocks, dc_format)
    if dc_codeblocks:
        discord_handler.setFormatter(color_formatter)
    else:
        discord_handler.setFormatter(logging.Formatter(dc_format))
    discord_handler.addFilter(_discord_filter)
    discord_handler.setLevel(discord_log_level)
    logger.addHandler(discord_handler)

    return logger



from __future__ import annotations

import re
from datetime import datetime, timedelta, timezone
from typing import Literal

from discord.utils import format_dt, utcnow

from .internal.translation import tp


def set_utc(dt: datetime) -> datetime:
    """Set the timezone of a datetime object to UTC.

    Parameters
    ----------
    dt:
        The datetime object to set the timezone of.

    Returns
    -------
    :class:`datetime.datetime`
    """
    return dt.replace(tzinfo=timezone.utc)


def convert_time(seconds: int | float, relative: bool = True) -> str:
    """Convert seconds to a human-readable time.

    Parameters
    ----------
    seconds:
        The amount of seconds to convert.
    relative: :class:`bool`
        Whether to use relative time. Defaults to ``True``.

        .. hint::
            This is only needed for German translation and will
            not have any effect if the language is set to English.

    Returns
    -------
    :class:`str`
        A human-readable time.
    """
    if seconds < 60:
        return f"{round(seconds)} {tp('sec', round(seconds))}"
    minutes = seconds / 60
    if minutes < 60:
        return f"{round(minutes)} {tp('min', round(minutes))}"
    hours = minutes / 60
    if hours < 24:
        return f"{round(hours)} {tp('hour', round(hours))}"
    days = hours / 24
    return f"{round(days)} {tp('day', round(days), relative=relative)}"


def convert_dt(dt: datetime | timedelta, relative: bool = True) -> str:
    """Convert :class:`datetime` or :class:`timedelta` to a human-readable time.

    This function calls :func:`convert_time`.

    Parameters
    ----------
    dt:
        The datetime or timedelta object to convert.
    relative: :class:`bool`
        Whether to use relative time. Defaults to ``True``.

    Returns
    -------
    :class:`str`
        A human-readable time.
    """
    if isinstance(dt, timedelta):
        return convert_time(abs(dt.total_seconds()), relative)

    if isinstance(dt, datetime):
        return convert_time(abs((dt - utcnow()).total_seconds()), relative)


def dc_timestamp(
    seconds: int | float, style: Literal["t", "T", "d", "D", "f", "F", "R"] = "R"
) -> str:
    """Convert seconds to a Discord timestamp.

    Parameters
    ----------
    seconds:
        The amount of seconds to convert.
    style: :class:`str`
        The style of the timestamp. Defaults to ``R``.
        For more information, see :func:`discord.utils.format_dt`.

    Returns
    -------
    :class:`str`
        A Discord timestamp.
    """
    dt = utcnow() + timedelta(seconds=seconds)
    return format_dt(dt, style)


def convert_to_seconds(s: str):
    """Convert a string to seconds.

    Parameters
    ----------
    s:
        The string to convert.

    Returns
    -------
    :class:`int`
        The amount of seconds.
    """
    units = {"s": "seconds", "m": "minutes", "h": "hours", "d": "days", "w": "weeks"}
    return int(
        timedelta(
            **{
                units.get(m.group("unit").lower(), "seconds"): float(m.group("val"))
                for m in re.finditer(r"(?P<val>\d+(\.\d+)?)(?P<unit>[smhdw]?)", s, flags=re.I)
            }
        ).total_seconds()
    )



import json
import os
from functools import cache
from pathlib import Path
from typing import Dict

from discord import Color, Embed

TEMPLATES: Dict[str, Embed] = {
    "success_embed": Embed(color=Color.green()),
    "error_embed": Embed(color=Color.red()),
    "warn_embed": Embed(color=Color.gold()),
    "info_embed": Embed(color=Color.blue()),
}


def save_embeds(**kwargs: Embed):
    """Save multiple embeds to a JSON file.

    If one of the default values is not included, a default template will be saved.
    """
    embeds = {}
    overrides = TEMPLATES if len(kwargs) == 0 else kwargs

    for name, embed in overrides.items():
        if embed is None:
            embeds[name] = TEMPLATES[name].to_dict()
        else:
            embeds[name] = embed.to_dict()

    parent = Path(__file__).parent.absolute()
    with open(os.path.join(parent, "embeds.json"), "w") as file:
        json.dump(embeds, file, indent=2)


@cache
def load_embed(name: str) -> Embed:
    """Load an embed template from a JSON file."""
    parent = Path(__file__).parent.absolute()
    json_path = parent.joinpath("embeds.json")
    if not json_path.exists():
        save_embeds()

    with open(os.path.join(parent, "embeds.json")) as file:
        embeds = json.load(file)

    try:
        return Embed.from_dict(embeds[name])
    except KeyError:
        if name in TEMPLATES.keys():
            save_embeds()
            return load_embed()
        else:
            raise ValueError(f"Embed template '{name}' not found.")



de = {
    "times": {
        "min": "Minute",
        "sec": "Sekunde",
        "hour": "Stunde",
        "day": "Tag",
    },
    "bot": {
        "error": "Der folgende Fehler ist aufgetreten",
        "cooldown": "Versuche es {} erneut.",
        "no_perm_title": "Fehlende Berechtigung",
        "no_perm_desc": "Mir fehlen die folgenden Berechtigungen, um diesen Befehl auszuführen.",
    },
}

en = {
    "times": {
        "min": "minute",
        "sec": "second",
        "hour": "hour",
        "day": "day",
    },
    "bot": {
        "error": "The following error occurred",
        "cooldown": "Try again {}.",
        "no_perm_title": "Missing permission",
        "no_perm_desc": "I'm missing the following permissions to execute this command.",
    },
}



from __future__ import annotations

import logging
import os
import traceback
import warnings
from pathlib import Path
from typing import Any, Literal

import aiohttp
import discord
from discord.ext import commands

from .emb import error as error_emb
from .internal.translation import set_lang, t
from .logs import DEFAULT_LOG, custom_log, set_log
from .times import dc_timestamp


class Bot(discord.Bot):
    """Bot class that extends from :class:`discord.Bot`.

    Parameters
    ----------
    intents:
        The intents to use for the bot. Defaults to :meth:`discord.Intents.default()`.
    debug:
        Enable debug logs. Defaults to ``True``.
    error_handler:
        Enable the error handler. Defaults to ``True``.
    error_webhook_url:
        The webhook URL to send error messages to. Defaults to ``None``.

        .. note::
            You need to enable the error handler for the webhook to work.
    ignored_errors:
        A list of error types to ignore. Defaults to ``None``.
    full_error_traceback:
        Whether to send the full error traceback. If this is ``False``,
        only the most recent traceback will be sent. Defaults to ``False``.
    language:
        The language to use for the bot. Defaults to ``en``.
    """

    def __init__(
        self,
        intents: discord.Intents = discord.Intents.default(),
        *,
        debug: bool = True,
        error_handler: bool = True,
        error_webhook_url: str | None = None,
        ignored_errors: list[Any] | None = None,
        full_error_traceback: bool = False,
        language: Literal["en", "de"] = "en",
        **kwargs,
    ):
        super().__init__(intents=intents, **kwargs)

        if debug:
            self.logger = set_log(DEFAULT_LOG)
        else:
            self.logger = logging.getLogger(DEFAULT_LOG)
            self.logger.addHandler(logging.NullHandler())

        self.error_webhook_url = error_webhook_url
        self.ignored_errors = ignored_errors or []
        self.full_error_traceback = full_error_traceback
        set_lang(language)

        if error_handler:
            self.add_listener(self._error_event, "on_application_command_error")
        elif error_webhook_url:
            warnings.warn("You need to enable the error handler for the webhook to work.")

        self.add_listener(self.ready_event, "on_ready")

    def load_cogs(
        self,
        *directories: str,
        subdirectories: bool = False,
        ignored_cogs: list[str] | None = None,
        custom_logs: bool | str = True,
    ):
        """Load all cogs in the given directories.

        Parameters
        ----------
        *directories:
            Names of the directories to load cogs from.
            Defaults to ``cogs``.
        subdirectories:
            Whether to load cogs from subdirectories.
            Defaults to ``False``.
        ignored_cogs:
            A list of cogs to ignore. Defaults to ``None``.
        custom_logs:
            Whether to use a custom log format for cogs. Defaults to ``True``.
            You can also pass in a custom color.
        """
        ignored_cogs = ignored_cogs or []
        if not directories:
            directories = ("cogs",)

        for directory in directories:
            path = Path(directory)

            for filename in os.listdir(directory):
                name = filename[:-3]
                if filename.endswith(".py") and name not in ignored_cogs:
                    self.load_extension(f"{'.'.join(path.parts)}.{name}")
                    if custom_logs:
                        custom_log("COG", f"Loaded {name}", color=custom_logs, level=logging.DEBUG)
                    else:
                        self.logger.debug(f"Loaded {name}")

            if subdirectories:
                for element in os.scandir(directory):
                    if not element.is_dir():
                        continue

                    for sub_file in os.scandir(element.path):
                        name = sub_file.name[:-3]
                        if sub_file.name.endswith(".py") and name not in ignored_cogs:
                            self.load_extension(f"{'.'.join(path.parts)}.{element.name}.{name}")
                            if custom_logs:
                                custom_log(
                                    "COG",
                                    f"Loaded {element.name}.{name}",
                                    color=custom_logs,
                                    level=logging.DEBUG,
                                )
                            else:
                                self.logger.debug(f"Loaded {element.name}.{name}")

    async def ready_event(self):
        """Prints the bot's information when it's ready."""
        infos = [
            f"User:     {self.user}",
            f"ID:       {self.user.id}",
            f"Pycord:   {discord.__version__}",
            f"Commands: {len(self.commands):,}",
            f"Guilds:   {len(self.guilds):,}",
            f"Latency:  {round(self.latency * 1000):,}ms",
        ]

        longest = max([str(i) for i in infos], key=len)
        formatter = f"<{len(longest)}"

        start_txt = "Bot is online!"
        start_txt += f"\n╔{(len(longest) + 2) * '═'}╗\n"
        for thing in infos:
            start_txt += f"║ {thing:{formatter}} ║\n"
        start_txt += f"╚{(len(longest) + 2) * '═'}╝"
        self.logger.info(start_txt)

    async def _error_event(self, ctx: discord.ApplicationContext, error: discord.DiscordException):
        """The event that handles application command errors."""
        if type(error) in self.ignored_errors:
            return

        if isinstance(error, commands.CommandOnCooldown):
            seconds = round(ctx.command.get_cooldown_retry_after(ctx))
            cooldown_txt = t("cooldown", dc_timestamp(seconds))
            await error_emb(ctx, cooldown_txt, title="Cooldown")

        elif isinstance(error, commands.BotMissingPermissions):
            perms = "\n".join(error.missing_permissions)
            perm_txt = f"{t('no_perm_desc')} ```\n{perms}```"
            await error_emb(ctx, perm_txt, title=t("no_perm_title"))

        else:
            if "original" in error.__dict__ and not self.full_error_traceback:
                original_error = error.__dict__["original"]
                error_msg = f"{original_error.__class__.__name__}: {error.__cause__}"
                error = original_error
            else:
                error_msg = f"{error}"

            error_txt = f"{t('error')}: ```{error_msg}```"
            try:
                await error_emb(ctx, error_txt, title="Error")
            except discord.HTTPException:
                pass

            webhook_sent = False
            if self.error_webhook_url:
                async with aiohttp.ClientSession() as session:
                    webhook = discord.Webhook.from_url(
                        self.error_webhook_url, session=session, bot_token=self.http.token
                    )
                    error_txt = "".join(
                        traceback.format_exception(type(error), error, error.__traceback__)
                    )
                    guild_txt = (
                        f"\n- **Guild:** {ctx.guild.name} - `{ctx.guild.id}`" if ctx.guild else ""
                    )
                    user_txt = (
                        f"\n- **User:** {ctx.author} - `{ctx.author.id}`" if ctx.author else ""
                    )

                    embed = discord.Embed(
                        title="Error Report",
                        description=f"- **Command:** /{ctx.command.qualified_name}"
                        f"{guild_txt}{user_txt}"
                        f"\n```py\n{error_txt[:3500]}```",
                        color=discord.Color.red(),
                    )
                    try:
                        await webhook.send(
                            embed=embed,
                            username=f"{self.user.name} Error Report",
                            avatar_url=self.user.display_avatar.url,
                        )
                    except discord.HTTPException:
                        self.logger.error(
                            "Error while sending error report to webhook. "
                            "Please check if the URL is correct."
                        )
                    else:
                        webhook_sent = True

            self.logger.exception(
                f"Error while executing **/{ctx.command.qualified_name}** ```{error_msg}```",
                exc_info=error,
                extra={"webhook_sent": webhook_sent},
            )



"""Internal language utilities for the library."""

import inspect
import os
from configparser import ConfigParser
from functools import cache
from pathlib import Path
from typing import Literal

from .languages import *


def plural_de(amount: int, word: str, relative: bool = True) -> str:
    """Pluralize a given word (German).

    Parameters
    ----------
    amount:
        The amount to check.
    word:
        The word to pluralize.
    relative:
        Whether to use relative time. Defaults to ``True``.

    Examples
    --------
    >>> plural_de(1, "Tag")
    'Tag'
    >>> plural_de(2, "Tag")  # Relative: "Seit 5 Tagen"
    'Tagen'
    >>> plural_de(2, "Tag", relative=False)  # Not relative: "5 Tage"
    'Tage'
    """
    if amount != 1:
        if not word.endswith("e"):
            word += "e"

        if relative:
            return f"{word}n"

    return word


def plural_en(amount: int, word: str) -> str:
    """Pluralize a given word (English).

    Parameters
    ----------
    amount:
        The amount to check.
    word:
        The word to pluralize.

    Returns
    -------
    :class:`str`
        The pluralized word.
    """
    if amount != 1:
        return f"{word}s"
    return word


def tp(key: str, amount: int, *args: str, relative: bool = True) -> str:
    """Load a string in the selected language and pluralize it.

    Parameters
    ----------
    key:
        The text to load.
    amount:
        The amount to check.
    *args:
        The arguments to format the string with.
    relative:
        Whether to use relative time. Defaults to ``True``.
    """
    word = t(key, *args)
    lang = get_lang()

    if lang == "de":
        return plural_de(amount, word, relative)
    else:
        return plural_en(amount, word)


def t(key: str, *args: str):
    """Load a string in the selected language.

    Parameters
    ----------
    key:
        The text to load.
    *args:
        The arguments to format the string with.
    """
    n = 1
    origin_file = Path(inspect.stack()[n].filename).stem

    while origin_file == Path(__file__).stem:
        n += 1
        origin_file = Path(inspect.stack()[n].filename).stem

    lang = get_lang()

    if lang == "de":
        return de[origin_file][key].format(*args)
    else:
        return en[origin_file][key].format(*args)


@cache
def get_lang():
    """Get the language from the config file."""
    parent = Path(__file__).parent.absolute()
    config = ConfigParser()
    config.read(os.path.join(parent, "config.ini"))

    return config["DEFAULT"]["lang"]


def set_lang(lang: Literal["en", "de"]):
    """Set the language in the config file.

    Parameters
    ----------
    lang:
        The language to set.
    """
    parent = Path(__file__).parent.absolute()
    config = ConfigParser()
    config["DEFAULT"] = {"lang": lang}

    with open(os.path.join(parent, "config.ini"), "w") as f:
        config.write(f)
